package controllers

import (

	"fmt"
	"time"
	"path"
	"strings"
	"strconv"
	"os/exec"
	"net/url"
	"net/http"
	"io/ioutil"
	"path/filepath"
	"encoding/json"
	"github.com/revel/revel"
)

type ImportCfg struct {

	 DomainToken string					`xml:"WebitelServers>Server>domainToken" json:"domainToken,omitempty"`
	 DialerID string					`xml:"WebitelServers>Server>dialerId" json:"dialerId,omitempty"`// >5911b9e5357471000cd63f9b</dialerId>
	 EngineURL string					`xml:"WebitelServers>Server>engineUrl" json:"engineUrl,omitempty"`// https://pre.webitel.com/engine/</engineUrl>
	 MembersAddUrn string				`xml:"WebitelServers>Server>dialerPostUrl" json:"membersAddUrn,omitempty"`// >api/v2/dialer/{0}/members/aggregate</membersLogPostUrl>

	 ConnectionString string			`xml:"WebitelServers>Server>connectionString" json:"connectionString,omitempty"`// >Data Source=PROJECT-DEV007;Initial Catalog=Webitel_CDR_Graphic_Monitoring;User ID=Supervisor</connectionString>
	 PreExecute string					`xml:"WebitelServers>Server>preExecuteSql,omitempty" json:"PreExecute,omitempty"`// SQLText
	 TableName string					`xml:"WebitelServers>Server>contactMembersTable" json:"SQLTable,omitempty"`// >Contact</contactMembersTable>

	 PrimaryColumnName string			`xml:"WebitelServers>Server>contactMembersPrimaryColumn" json:"PrimaryColumn"`// >Id</contactMembersPrimaryColumn>
	 CustomFilter string				`xml:"WebitelServers>Server>contactMembersSQLFilter" json:"SQLFilter"`//>

	 ExportDateTimeColumnName string	`xml:"WebitelServers>Server>importDateMembersColumn" json:"DataTime"` // >UsrImportDate</importDateMembersColumn>
	 ExpiresIn string					`xml:"WebitelServers>Server>membersTTL,omitempty" json:"expires"`// >10h</membersTTL>
	 
	 TableMap TableMap					`xml:"ContactMembersTableFields" json:"ColumnMappings,omitempty"`
}

// Validate as complete configuration for import operation to perform.
func (cfg *ImportCfg) Validate(v *revel.Validation) {
	
	v.Required(cfg.DomainToken)
	v.Required(cfg.EngineURL)
	v.Required(cfg.DialerID)
	
	v.Required(cfg.ConnectionString)
	v.Required(cfg.TableName)
	v.Required(cfg.PrimaryColumnName)
	
	// app.conf: import.members.path =
	if cfg.MembersAddUrn == "" {
		cfg.MembersAddUrn = revel.Config.
			StringDefault("import.members.urn", "/api/v2/dialer/{0}/members")
	}
	cfg.MembersAddUrn = strings.Replace(cfg.MembersAddUrn,
		"/"+ cfg.DialerID +"/", "/{0}/", -1)
	
	v.Required(cfg.MembersAddUrn)

	v.Required(cfg.TableMap)
}

type ImportReq struct {
	 Profile	ImportCfg				`json:"sql,omitempty"`
	 Params		map[string]interface{}	`json:"webitel,omitempty"`
}

/*func (c App) ImportDone() revel.Result {
	body := c.Request.GetBody()
	dump, _ := ioutil.ReadAll(body)
	c.Log.Debugf("\n\n%s\n\n", dump)
	return nil // c.RenderText("OK")
}*/

// POST /import
func (c App) ImportMembers() revel.Result {
	
	if len(c.Params.JSON) == 0 {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Error(c.Controller, status, "sql4dialer/import: configuration is missing")
	}

	var (
		req ImportReq
		resultURL string
		resultHost string
		resultPath string
	)
	// Bind some HTTP.Header arguments
	for header, bind := range map[string]*string {
	//	"X-Action": nil,
	//	"X-Template-Id": nil,
		"X-Dialer-Id": &req.Profile.DialerID,
	//	"X-Domain-Token": &cfg.DomainToken,
		"X-Response-URL": &resultURL,
		"X-Response-Host": &resultHost,
		"X-Response-Path": &resultPath,
	} {
		*bind = c.Request.GetHttpHeader(header)
	}

	// Bind the JSON data to the config.
	err := json.Unmarshal(c.Params.JSON, &req)
	if err != nil {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Result{ Code:status, Data:err.Error(), Text:"sql4dialer/import: configuration error" }
		// return Error(c.Controller, status, "sql4dialer/import: configuration error")
	}
	// X-Response-URL: X-Response-Host[X-Response-Path]
	if resultURL == "" {
		url, err := url.Parse(resultHost)
		if err != nil {
			c.Validation.Errors = append(c.Validation.Errors, &revel.ValidationError{
				Key: "X-Response-Host", Message: err.Error(),
			})
		} else if resultPath != "" {
			url.Path = path.Join(url.Path, resultPath)
		}
		resultURL = url.String()
	}
	// <engineUrl/> == X-Response-Host
	if req.Profile.EngineURL == "" {
		req.Profile.EngineURL = resultHost
	}
	// <domainToken> == X-Access-Token
	if req.Profile.DomainToken == "" {
		req.Profile.DomainToken, _ =
			req.Params["token"].(string)
	}

	if req.Profile.ExpiresIn == "" {
		req.Profile.ExpiresIn, _ =
			req.Params["expire"].(string)
	}

	c.Validation.Required(resultURL)
	// TODO: Validate(config)
	req.Profile.Validate(c.Validation)
	if c.Validation.HasErrors() {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Result{ Code:status, Data:c.Validation.Errors }
		// // Store the validation errors in the flash context and redirect.
		// c.Validation.Keep()
		// c.FlashParams()
		// return c.Redirect(Hotels.Settings)
	}

	pid, err := ImportStart(req.Profile, resultURL)
	if err != nil {
		return Error(c.Controller, 1, err.Error())
	}

	return c.ProcessStatus(pid)
}

func ImportStart(cfg ImportCfg, resultURL string) (pid int, err error) {
	// NewImport()

	now := time.Now()
	oid := strconv.FormatInt(now.UnixNano(), 10)

	task := new(Proc)
	// task.cfg = cfg
	task.ResultURL = resultURL
	task.LogFile = oid +".import.log"
	task.CfgFile = oid+ ".import.cfg"
	revel.AppLog.Debugf("Storing Config.xml: %s", task.CfgFile)
	cfgPath := filepath.Join(RunPath, task.CfgFile)
	// Marshal XML
	var xdat []byte
	xdat, err = ConfigXML(cfg) // xml.MarshalIndent(cfg, "", "\t")
	if err != nil {
		return // 0, err
	}
	// Save File
	err = ioutil.WriteFile(cfgPath, xdat, 755)
	if err != nil {
		return // 0, err
	}

	task.cmd = exec.Command(filepath.Join(BinPath, ExeFile), "/dialer", fmt.Sprintf(`/conf=%s`, cfgPath),
		fmt.Sprintf(`/log=%s`, filepath.Join(RunPath, task.LogFile)))
	
	task.cmd.Stdout = &task.LastLog // &task.capture
	task.cmd.Stderr = &task.LastErr // &task.capture

	err = task.cmd.Start()
	if err != nil {
		revel.AppLog.Errorf("Failed: %v", err)
		return // 0, err
	}
	// TODO: map running
	pid = task.cmd.Process.Pid
	revel.AppLog.Debugf(`Task[%d]: start %s`, pid, task.CfgFile)

	task.Pid = task.cmd.Process.Pid
	task.StartedAt = now // time.Now()
	task.tokenAuth = cfg.DomainToken
	
	taskMu.Lock()
	tasks[pid] = task
	taskMu.Unlock()

	go complete(task, nil) // async
	return // pid, nil // +OK started
}