package controllers

import (

	"time"
	"sync"
	"bytes"
	"regexp"
	"os/exec"
	"net/url"
	"net/http"
	"io/ioutil"
	"encoding/json"
	"github.com/revel/revel"
)


type Proc struct {

	 Pid		int			`json:"processId,omitempty"`
	 StartedAt	time.Time	`json:"startedAt,omitempty"`
	 FinishedAt	*time.Time	`json:"finishedAt,omitempty"`
	 Error		string		`json:"exitError,omitempty"`
	 // Args
	 CfgFile	string		`json:"cfgFile,omitempty"`
	 LogFile	string		`json:"logFile,omitempty"`
	 LastLog	stdLine		`json:"lastLog,omitempty"`
	 LastErr	stdLine		`json:"lastErr,omitempty"`
	 
	 ResultURL	string		`json:"resultUrl,omitempty"`
	 tokenAuth	string	//	`json:"-"`

	 cmd		*exec.Cmd // runtime.process
}

// func (e *Proc) Start(args ...string) (pid int, err error) {

// }

// func (e *Proc) Complete(respondURL string, respondData func(*Proc) interface{}) {
	
// }


// GET /
func (c App) ProcessTasks() revel.Result {
	taskMu.Lock()
	rtm := make([]*Proc, 0, len(tasks))
	for _, e := range tasks {
		rtm = append(rtm, e)
	}
	taskMu.Unlock()
	return Result{ Data: rtm }
	// return Result{ Data: []interface{}(nil) }
}

// GET /:pid
func (c App) ProcessStatus(pid int) revel.Result {
	
	taskMu.Lock()
	task, found := tasks[pid]
	// _, found := tasks[pid]
	taskMu.Unlock()

	if !found {
		status := http.StatusNotFound
		c.Response.Status = status
		return Result{ Code:status, Text:http.StatusText(status) }
	}
	
	// return Result{ Text: fmt.Sprintf("%s", task.ProcessState) } // Data:task }
	return Result{ Data:task } // { Text:"+OK" }
	// return Result{ Text:"+OK" }
}

var (
	taskMu sync.Mutex
	tasks  = make(map[int]*Proc)
)

func complete(task *Proc, resultCook func(*Proc) interface{}) {
	
	exitErr := task.cmd.Wait()
	finishAt := time.Now()
	// Done
	taskMu.Lock()
	if tasks[task.Pid] == task {
		delete(tasks, task.Pid)
	}
	taskMu.Unlock()
	task.FinishedAt = &finishAt

	var (
		statusOk bool // false
		statusDescr string
	)

	if exitErr != nil {
		task.Error = exitErr.Error()
		revel.AppLog.Errorf("Task[%d] Process exit with error: %v", task.Pid, exitErr)
		// statusOk = false
		statusDescr = "Process exit: "+ task.Error // "Process exit" 
	} else {
		task.Error = task.LastErr.Message()
		if task.Error != "" {
			// statusOk = false
			statusDescr = "Complete with errors: "+ task.Error
			revel.AppLog.Errorf("Task[%d] Complete with error: %s", task.Pid, statusDescr)
		} else {
			statusOk = true
			statusDescr = "Complete: "+ task.LastLog.Message()
			revel.AppLog.Infof("Task[%d] Complete successfully! %s", task.Pid, statusDescr)
		}
	}

	resultOk := "false"
	if statusOk {
		resultOk = "true"
	}

	var (
		resultJSON []byte
		resultData interface{}
		// resultBody func() (io.ReadCloser, error)
	)
	
	if resultCook != nil {
		resultData = resultCook(task)
	}/* else { // DEBUG !!!
		resultData = task
	}*/

	var err error
	if resultData != nil {
		resultJSON, err = json.MarshalIndent(resultData, "", "\t")
		if err != nil {
			panic(err)
		}
		// resultBody = func() (io.ReadCloser, error) {
		// 	r := bytes.NewReader(resultJSON)
		// 	return ioutil.NopCloser(r), nil
		// }
	}
	
	resultURL := task.ResultURL +"?success="+ resultOk +"&message="+ url.QueryEscape(statusDescr)
	revel.AppLog.Infof("Task[%d] Responds to %s\n%s", task.Pid, resultURL, resultJSON)
	resultReq, err := http.NewRequest(http.MethodPut, resultURL, nil)
	if err != nil {
		panic(err)
	}
	resultReq.Header.Set("User-Agent", "sql4dialer/v0.1")
	resultReq.Header.Set("Content-Type", "application/json; charset=utf-8")
	resultReq.Header.Set("X-Access-Token", task.tokenAuth)
	// if resultBody != nil {
	// 	resultReq.GetBody = resultBody
	// 	resultReq.ContentLength = int64(len(resultJSON))
	// }

	attemptN := 1
	attemptsMax := 3
	fallbackDelay := time.Second *5
	for { // attemptN := 1; attemptN < attemptsMax; attemptN++ {
		
		resultReq.ContentLength = int64(len(resultJSON))
		if resultReq.ContentLength != 0 {
			resultReq.Body = ioutil.NopCloser(bytes.NewReader(resultJSON))
		}
		
		resp, respErr := http.DefaultClient.Do(resultReq)
		// resp, respErr := http.Post(task.ResultURL, "application/json; charset=utf-8", bytes.NewReader(jdat)) // &task.capture)
		if respErr == nil && resp.StatusCode == http.StatusOK {
			revel.AppLog.Debugf("Task[%d]: StatusDelivery(attempt:%d): <<== %d (%s)", task.Pid, attemptN, resp.StatusCode, resp.Status)
			break // Success: 200 (OK)
		}

		if respErr != nil {
			revel.AppLog.Errorf("Task[%d]: StatusDelivery(attempt:%d): error: %s", task.Pid, attemptN, respErr)
		} else if resp.StatusCode != http.StatusOK {
			revel.AppLog.Warnf("Task[%d]: StatusDelivery(attempt:%d): <<== %s", task.Pid, attemptN, resp.Status)
		}

		attemptN++
		if attemptN > attemptsMax {
			revel.AppLog.Debugf("Task[%d]: StatusDelivery: reach max %d attempts; mark as done !", task.Pid, attemptsMax)
			break
		}

		revel.AppLog.Debugf("Task[%d]: StatusDelivery: next attempt %d in %s", task.Pid, attemptN, time.Duration(fallbackDelay))
		time.Sleep(fallbackDelay)
		// continue
	}

	revel.AppLog.Debugf("Task[%d]: Operation done!", task.Pid)
	// DONE: thread
}

// stdLine implements io.Writer that grab only last continued line.
type stdLine struct {
	 line bytes.Buffer
	 last byte
}

func (w *stdLine) String() string {
	return w.line.String()
}

// Write writes/holds only last line ...
func (w *stdLine) Write(p []byte) (n int, err error) {
	n = len(p)
	if n == 0 {
		return // 0, nil
	}
	if w.last == '\n' {
		w.line.Reset()
		w.last = 0
	}
	// lineFeed
	switch lf := bytes.LastIndexByte(p,'\n'); lf {
	case n-1 : // last byte
	{	// lineStart
		ls := bytes.LastIndexByte(p[0:n-1],'\n')
		w.line.Write(p[ls+1:])
		w.last = p[n-1]
	}
	default: // inner byte
	{
		w.line.Write(p[lf+1:])
		w.last = p[n-1]
	}
	case -1 : // not found
	{
		w.line.Write(p)
		w.last = p[n-1]
	}}
	
	// revel.AppLog.Infof("%q", p)
	return // len(p), nil
}

// MarshalText implements here for JSON marshaling
func (w *stdLine) MarshalText() ([]byte, error) {
	return dropCRLF(w.line.Bytes()), nil
}

// 07.12.2017 12:26:47 [INFO] ${message}[\r\n]
var logRegular = regexp.MustCompile(`\[(INFO|WARN|ERR )\]\s+([^\r\n]*)`)

func (w *stdLine) Message() string {
	text := w.line.Bytes()
	if len(text) == 0 {
		return ""
	}
	i := 2 // match: index
	// NOTE:  match: 0 = input; 1 .. = brackets
	mi := logRegular.FindSubmatchIndex(text)
	// match: found ?
	if (2*i+1) < len(mi) && mi[2*i] != -1 {
		return string(text[mi[2*i]:mi[2*i+1]])
	}
	return string(text)
}

// dropCR drops a terminal [\n]\r from the data.
func dropCRLF(data []byte) []byte {
	var c byte
	for len(data) > 0 {
		c = data[len(data)-1]
		if c != '\n' && c != '\r' {
			break
		}
		data = data[0:len(data)-1]
	}
	return data
}


// type logLast struct {
// 	 Tail int // lines count to buffer in
// 	 buf bytes.Buffer
// 	 lfp [ ]int
// }

// // Write writes/holds only last line ...
// func (w *logLast) Write(p []byte) (n int, err error) {
// 	n = len(p)
// 	if n == 0 {
// 		return // 0, nil
// 	}
// 	// if w.last == '\n' {
// 	// 	w.line.Reset()
// 	// 	w.last = 0
// 	// }
// 	// lineFeed



// 	lfn := bytes.Count(p, []byte{'\n'})
// 	if lfn == 0 {
// 		w.buf.Write(p)
// 		return
// 	}
// 	// Jitter Flush
// 	if len(w.lfp) +lfn > w.Tail {
// 		fln := (len(w.lfp)+lfn)-w.Tail
// 		w.buf.Truncate(w.lfp[fln])

// 	}

// 	for len(p) > 0 {

// 		lfp := bytes.IndexByte(p,'\n')
// 		if lfp == -1 { // not[ more] found
// 			w.buf.Write(p)
// 			break
// 		}
// 		w.buf.Write(p[0:lfp+1])
// 		w.lfp = append(w.lfp, w.buf.Len()-1)
// 		p = p[lfp+1:]
// 	}


// 	switch lf := bytes.LastIndexByte(p,'\n'); lf {
// 	case n-1 : // last byte
// 	{	// lineStart
// 		ls := bytes.LastIndexByte(p[0:n-1],'\n')
// 		w.buf.Write(p[ls+1:])
// 		w.last = p[n-1]
// 	}
// 	default: // inner byte
// 	{
// 		w.buf.Write(p[lf+1:])
// 		w.last = p[n-1]
// 	}
// 	case -1 : // not found
// 	{
// 		w.buf.Write(p)
// 		w.last = p[n-1]
// 	}}
	
// 	// revel.AppLog.Infof("%q", p)
// 	return // len(p), nil
// }