package controllers

import (

	"fmt"
	"bytes"
	"encoding/xml"
	"encoding/json"
)


func ConfigXML(cfg interface{}) ([]byte, error) {
	
	var (
		err error
		buf bytes.Buffer
	)

	xwr := xml.NewEncoder(&buf)
	xwr.Indent("", "\t")

	for _, token := range [ ]xml.Token {
		xml.ProcInst{ Target:"xml", Inst:[]byte("version=\"1.0\" encoding=\"UTF-8\"")},
		xml.CharData{ '\n' },
	}{
		err = xwr.EncodeToken(token)
		if err != nil {
			return nil, err
		}
	}

	err = xwr.EncodeElement( cfg,
		xml.StartElement{ Name:xml.Name{ Local: "Config" }, Attr: [ ]xml.Attr {
			{ Name:xml.Name{ Local: "xmlns:xsi" }, Value: "http://www.w3.org/2001/XMLSchema-instance" },
			{ Name:xml.Name{ Local: "xmlns:xsd" }, Value: "http://www.w3.org/2001/XMLSchema" },
		}},
	)
	if err != nil {
		return nil, err
	}

	return buf.Bytes(), nil
}


type Parameter struct { Name, Value string }

func (p Parameter) MarshalXML(xwr *xml.Encoder, _ xml.StartElement) error {
	if p.Name == "" {
		return nil
	}
	return xwr.EncodeElement(p.Value, xml.StartElement{Name:xml.Name{Local:p.Name}})
}

func (p *Parameter) UnmarshalXML(xrd *xml.Decoder, _ xml.StartElement) error {
	var (
		elem xml.StartElement
		innerText string
	)
	err := xrd.DecodeElement(&innerText, &elem)
	if err != nil {
		return err
	}
	p.Name = elem.Name.Local
	p.Value = innerText
	return nil
}



type Parameters [ ]Parameter

func (p Parameters) MarshalXML(xwr *xml.Encoder, _ xml.StartElement) error {
	// fmt.Printf("Contact2MembersTableMap(%s) %v\n", xmap, m)
	
	// err := xwr.EncodeToken(xcol)
	// if err != nil {
	// 	return err
	// }

	for _, x := range p {
		err := xwr.Encode(x)
		if err != nil {
			return err
		}
	}
	
	return nil // xwr.EncodeToken(xcol.End())
}

func (p *Parameters) UnmarshalXML(xrd *xml.Decoder, xcol xml.StartElement) error {
	var (
		
		err error
		elem xml.Name
		innerText string
		token xml.Token
		coll = make(Parameters, 0, 8)
	)
	
	for {

		token, err = xrd.Token()
		if err != nil {
			return err
		}

		switch unit := token.( type ) {
		case xml.StartElement:
		{
			if elem.Local == "" {
				elem = unit.Name
				break //switch
			}
			return fmt.Errorf("sql4dialer/xml: invalid element %s at position %d near <%s/>",
				unit, xrd.InputOffset(), xcol)
		}
		case xml.EndElement:
		{
			if elem.Local == unit.Name.Local && elem.Space == unit.Name.Space {
				coll = append(coll, Parameter{ Name:elem.Local, Value:innerText })
				elem.Local = ""
				elem.Space = ""
				innerText = ""
				break // switch
			}
			if xcol.Name.Local == unit.Name.Local && xcol.Name.Space == unit.Name.Space {
				*p = coll
				return nil 
			}
			return fmt.Errorf("sql4dialer/xml: invalid element %s at position %d near <%s/>",
				innerText, xrd.InputOffset(), xcol)
		}
		case xml.CharData:
		{
			for b := 0; b < len(unit); b++ {
				if !whitespace(unit[b]){
					innerText = string(unit)
					break // switch
				}
			}
		}
		default:
			return fmt.Errorf("sql4dialer/xml: invalid node %s at position %d near <%s/>",
				innerText, xrd.InputOffset(), xcol)
		}
	}
}

func (p *Parameters) UnmarshalJSON(data []byte) error {
	// fmt.Printf("Contact2MembersTableMap(%s)\n", data)
	
	var (
		
		key string
		buf = bytes.NewReader(data)
		jrd = json.NewDecoder(buf)
		coll = make(Parameters, 0, 8)
	)

	token, err := jrd.Token()
	if err != nil {
		return err
	}
	jsonStartObject, ok := token.(json.Delim)
	if !ok || jsonStartObject != '{' {
		return fmt.Errorf("sql4dialer/json: expect JSON object for {<parameters>}")
	}

	for {

		token, err = jrd.Token()
		if err != nil {
			return err
		}

		switch token.( type ) {
		case json.Delim:
		{
			if token.(json.Delim) == '}' { // jsonEndObject
				if *(p) = coll; !jrd.More() {
					return nil
				}
				// jrd.Buffered()
				return fmt.Errorf("sql4dialer/json: unexpected data at the end of {<parameters>}")
			}
			return fmt.Errorf("sql4dialer/json: unexpected token %q for {<parameters>}", token)
		}
		case string:
		{
			if key == "" {
				key = token.(string)
				break // switch
			}
			// tbl.Add(key, token.(string))
			coll = append(coll, Parameter{ Name:key, Value:token.(string) })
			// tbl[key] = append(tbl[key], token.(string))
			key = "" //
		}
		default:
		{
			return fmt.Errorf("sql4dialer/json: invalid token %q for {<parameters>}", token)
		}}

		// fmt.Printf("Json: %#v; Err: %v\n", token, err)
	}
}



type TableMap struct { // [ ]Parameter
	 Parameters // implements
}

func (m TableMap) MarshalXML(xwr *xml.Encoder, xcol xml.StartElement) error {
	// fmt.Printf("Contact2MembersTableMap(%s) %v\n", xmap, m)
	
	err := xwr.EncodeToken(xcol)
	if err != nil {
		return err
	}

	err = m.Parameters.MarshalXML(xwr, xcol)
	if err != nil {
		return err
	}
	
	return xwr.EncodeToken(xcol.End())
}





func whitespace(r byte) bool {
	switch r {
	case '\t', '\n', '\v', '\f', '\r', ' ', 0x85, 0xA0:
		return true
	}
	return false
}