package controllers

import (

	"fmt"
	"time"
	"path"
	"os/exec"
	"path/filepath"
	"strconv"
	"strings"
	"net/url"
	"net/http"
	"io/ioutil"
	"encoding/xml"
	"encoding/json"
	"github.com/revel/revel"
)

// REQUIRED
// ------------------------
// connectionString
// !lastStatisticTimestamp
// ------------------------
// engineUrl
// domainToken
// dialerId
// membersLogPostUrl

type ExportCfg struct {

	 DomainToken string					`xml:"WebitelServers>Server>domainToken" json:"domainToken,omitempty"`
	 EngineURL string					`xml:"WebitelServers>Server>engineUrl" json:"engineUrl,omitempty"`// https://pre.webitel.com/engine/</engineUrl>
	 DialerID string					`xml:"WebitelServers>Server>dialerId" json:"dialerId,omitempty"`// >5911b9e5357471000cd63f9b</dialerId>

	 ConnectionString string			`xml:"WebitelServers>Server>connectionString" json:"connectionString,omitempty"`// >Data Source=PROJECT-DEV007;Initial Catalog=Webitel_CDR_Graphic_Monitoring;User ID=Supervisor</connectionString>
	 PreExecute string					`xml:"WebitelServers>Server>preExecuteSql,omitempty" json:"PreExecute,omitempty"`// SQLText
	 TableName string					`xml:"WebitelServers>Server>contactStatisticTable" json:"SQLTable,omitempty"`// >Contact</contactMembersTable>

	 MembersLogUrn string				`xml:"WebitelServers>Server>membersLogPostUrl" json:"membersLogUrn,omitempty"`// >api/v2/dialer/{0}/members/aggregate</membersLogPostUrl>
	 LastStatisticTimestamp int64		`xml:"WebitelServers>Server>lastStatisticTimestamp" json:"lastStatisticTimestamp,omitempty"`// ></lastStatisticTimestamp>

	 Parameters [ ]Parameter			`xml:"WebitelServers>Server>parameters,omitempty" json:"-"` // 
	 TableMap TableMap					`xml:"ContactStatisticTableFields" json:"ColumnMappings,omitempty"`
}

type ExportReq struct {
	 Profile	ExportCfg				`json:"sql,omitempty"`
	 Params		map[string]interface{}	`json:"webitel,omitempty"`
	 Control	ExportControl			`json:"user_data,omitempty"`
}

// Validate as complete configuration for import operation to perform.
func (cfg *ExportCfg) Validate(v *revel.Validation) {

	v.Required(cfg.DomainToken)
	
	v.Required(cfg.ConnectionString)
	v.Required(cfg.TableName)

	v.Required(cfg.EngineURL)
	v.Required(cfg.DialerID)
	// app.conf: export.members.path =
	if cfg.MembersLogUrn == "" {
		cfg.MembersLogUrn = revel.Config.
			StringDefault("export.members.urn",
				"/api/v2/dialer/{0}/members/aggregate")
	}
	cfg.MembersLogUrn = strings.Replace(cfg.MembersLogUrn,
		cfg.DialerID, "{0}", -1)
	
	v.Required(cfg.MembersLogUrn)

	v.Required(cfg.TableMap)
}



// POST /export
func (c App) ExportMembersLog() revel.Result {
	
	if len(c.Params.JSON) == 0 {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Error(c.Controller, status, "sql4dialer/import: configuration is missing")
	}

	var (
		req ExportReq
		resultURL string
		resultHost string
		resultPath string
	)
	// Bind some HTTP.Header arguments
	for header, bind := range map[string]*string {
	//	"X-Action": nil,
	//	"X-Template-Id": nil,
		"X-Dialer-Id": &req.Profile.DialerID,
	//	"X-Domain-Token": &req.DialerSrv.DomainToken,
		"X-Response-URL": &resultURL,
		"X-Response-Host": &resultHost,
		"X-Response-Path": &resultPath,
	} {
		*bind = c.Request.GetHttpHeader(header)
	}

	// Bind the JSON data to the config.
	err := json.Unmarshal(c.Params.JSON, &req)
	if err != nil {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Result{ Code:status, Data:err.Error(), Text:"sql4dialer/export: configuration error" }
		// return Error(c.Controller, status, "sql4dialer/import: configuration error")
	}
	// X-Response-URL: X-Response-Host[X-Response-Path]
	if resultURL == "" {
		url, err := url.Parse(resultHost)
		if err != nil {
			c.Validation.Errors = append(c.Validation.Errors, &revel.ValidationError{
				Key: "X-Response-Host", Message: err.Error(),
			})
		} else if resultPath != "" {
			url.Path = path.Join(url.Path, resultPath)
		}
		resultURL = url.String()
	}
	// <engineUrl/> == X-Response-Host
	if req.Profile.EngineURL == "" {
		req.Profile.EngineURL = resultHost
	}
	// <domainToken> == X-Access-Token
	if req.Profile.DomainToken == "" {
		req.Profile.DomainToken, _ =
			req.Params["token"].(string)
	}
	// <lastStatisticTimestamp/>
	if req.Profile.LastStatisticTimestamp == 0 {
		req.Profile.LastStatisticTimestamp = req.Control.LastExportTimestamp
	}

	c.Validation.Required(resultURL)
	// TODO: Validate(config)
	req.Profile.Validate(c.Validation)
	if c.Validation.HasErrors() {
		status := http.StatusBadRequest
		c.Response.Status = status // 400 (Bad Request)
		return Result{ Code:status, Data:c.Validation.Errors }
		// // Store the validation errors in the flash context and redirect.
		// c.Validation.Keep()
		// c.FlashParams()
		// return c.Redirect(Hotels.Settings)
	}
	// Map Parameters
	for i, p := range req.Profile.TableMap.Parameters {
		cn := len(p.Value)
		if cn > 2 && p.Value[0] == '`' && p.Value[cn-1] == '`' {
			req.Profile.Parameters = append(req.Profile.Parameters,
				Parameter{ Name:"parameter"+strconv.Itoa(len(req.Profile.Parameters)), Value:strings.Trim(p.Value, "`") })
			req.Profile.TableMap.Parameters[i].Value = `%%`+ req.Profile.Parameters[len(req.Profile.Parameters)-1].Name +`%%`
		}
	} 

	pid, err := ExportStart(req.Profile, resultURL)
	if err != nil {
		return Error(c.Controller, 1, err.Error())
	}

	return c.ProcessStatus(pid)
}



func ExportStart(cfg ExportCfg, resultURL string) (pid int, err error) {
	// NewImport()

	now := time.Now()
	oid := strconv.FormatInt(now.UnixNano(), 10)

	task := new(Proc)
	// task.cfg = config
	task.ResultURL = resultURL
	task.LogFile = oid +".export.log"
	task.CfgFile = oid +".export.cfg"
	revel.AppLog.Debugf("Storing Config.xml: %s", task.CfgFile)
	cfgPath := filepath.Join(RunPath, task.CfgFile)
	// Marshal XML
	var xdat []byte
	xdat, err = ConfigXML(cfg) // xml.MarshalIndent(cfg, "", "\t")
	if err != nil {
		return // 0, err
	}
	// Save File
	err = ioutil.WriteFile(cfgPath, xdat, 755)
	if err != nil {
		return // 0, err
	}

	task.cmd = exec.Command(filepath.Join(BinPath, ExeFile), "/members_log", fmt.Sprintf(`/conf=%s`, cfgPath),
		fmt.Sprintf(`/log=%s`, filepath.Join(RunPath, task.LogFile)))
	
	task.cmd.Stdout = &task.LastLog
	task.cmd.Stderr = &task.LastErr

	err = task.cmd.Start()
	if err != nil {
		revel.AppLog.Errorf("Failed: %v", err)
		return // 0, err
	}
	// TODO: map running
	pid = task.cmd.Process.Pid
	revel.AppLog.Debugf(`Task[%d]: start %s`, pid, task.CfgFile)

	task.Pid = task.cmd.Process.Pid
	task.StartedAt = now // time.Now()
	task.tokenAuth = cfg.DomainToken
	
	taskMu.Lock()
	tasks[pid] = task
	taskMu.Unlock()

	go complete(task, controlExportResult) // async
	return // pid, nil // +OK started
}

// ExportControl Arguments
type ExportControl struct {
	 LastExportTimestamp int64 `xml:"WebitelServers>Server>lastStatisticTimestamp,omitempty" json:"lastStatisticTimestamp,omitempty"`
}

func controlExportResult(task *Proc) interface{} {
	
	xdat, err := ioutil.ReadFile(path.Join(RunPath, task.CfgFile))
	if err != nil {
		panic(err)
	}

	var res ExportControl
	err = xml.Unmarshal(xdat, &res)
	if err != nil {
		panic(err)
	}

	// revel.AppLog.Debugf("%#v", ExportResult)
	if res.LastExportTimestamp == 0 {
		return nil
	}
	
	return res
}