package controllers

import (
	"fmt"
	"net/http"
//	"encoding/xml"
	"encoding/json"
	"github.com/revel/revel"
)

func Error(c *revel.Controller, code int, text string, args ...interface{}) revel.Result {
	if code == 0 {
		code = 1
	}
	if text == "" {
		if len(args) != 0 {
			text = fmt.Sprint(args...)
		}
	} else if len(args) != 0 {
		text = fmt.Sprintf(text, args...)
	}
	return Result {
		Code: code,
		Text: text,
	}
}

type Result struct {
	 Code	int				`json:"status,omitempty" xml:"code,attr"`
	 Data	interface{}		`json:"result,omitempty" xml:"result,omitempty"`
	 Text	string			`json:"message,omitempty" xml:"message,omitempty"`
}

func (res Result) Apply(req *revel.Request, resp *revel.Response) {
	var (
		err error
		buf []byte
	)

	if revel.Config.BoolDefault("results.pretty", false) {
		buf, err = json.MarshalIndent(&res, "", "\t")
	} else {
		buf, err = json.Marshal(&res)
	}

	if err != nil {
		resp.Status = http.StatusInternalServerError
		Result{
			Code: http.StatusInternalServerError,
			Text: "result: could not marshal JSON",
			Data: err.Error(),
		}.Apply(req, resp);
		return
	}

	resp.WriteHeader(http.StatusOK, "application/json; charset=utf-8")
	if _, err = resp.GetWriter().Write(buf); err != nil {
		revel.AppLog.Error("Result: Response write failed:", "error", err)
	}
	return
}